
/** @file
    @brief		蔻NX
    @author		t`
    @since		2003.07.01
*/

//======================================================================
// INCLUDE
//======================================================================
#include "LunaBase.h"

#include "LunaCollision.h"

#include "LunaVector2.h"
#include "LunaVector3.h"
#include "LunaMatrix.h"
#include "LunaQuaternion.h"
#include "LunaStyle.h"

namespace lunadxmath{
//======================================================================
// STRUCT
//======================================================================

/**
	@brief `

	LunaCollisionNXŎgpĂ
	`\̂łB
*/
typedef struct LINE3D2
{
	struct {
		Float a;
		Float b;
	} x, y, z;
}
LINE3D2, *LPLINE3D2;


//===================================================================================
//===================================================================================
//
//
//			Qc֘A
//
//
//===================================================================================
//===================================================================================



//=============================================================================
/**
	@brief _(x,y) l ̂ǂ瑤ɂ邩ׂ

	@param x		[in] ]_̂wW
	@param y		[in] ]_̂xW
	@param pLine	[in] ]

	@retval > 0		 l ̉Eɂ
	@retval ==0		 l ̒ɂ
	@retval < 0		 l ̍ɂ

*/
//=============================================================================
static inline Float Side( Float x, Float y, LINE2D *pLine )
{
	return ((x - pLine->Pt0.Px) * (y - pLine->Pt1.Py) - (y - pLine->Pt0.Py) * (x - pLine->Pt1.Px));
}

//=============================================================================
/**
	@brief ̃xNg

	@param pDst		[out] xNgi[
	@param pSrc		[in] ɂȂ钼
*/
//=============================================================================
static inline void CreateLine2( LINE3D2 *pDst, LINE3D *pSrc )
{
	pDst->x.a = pSrc->Pt0.Px;
	pDst->x.b = pSrc->Pt1.Px - pSrc->Pt0.Px;

	pDst->y.a = pSrc->Pt0.Py;
	pDst->y.b = pSrc->Pt1.Py - pSrc->Pt0.Py;

	pDst->z.a = pSrc->Pt0.Pz;
	pDst->z.b = pSrc->Pt1.Pz - pSrc->Pt0.Pz;
}

//=============================================================================
/**
	_ƒ̓蔻s܂

	@param pPt1		[in] ]Qcʏ̓_
	@param pPt2		[in] ]Qcʏ̓_

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Point( POINT2D *pPt1, POINT2D *pPt2 )
{
	CIRCLE2D Cir1 = { pPt1->Px, pPt1->Py, 1.0f };
	CIRCLE2D Cir2 = { pPt2->Px, pPt2->Py, 1.0f };

	return Circle_Circle( &Cir1, &Cir2 );
}

//=============================================================================
/**
	_ƒ̓蔻s܂

	@param pPt		[in] ]Qcʏ̓_
	@param pLine	[in] ]Qcʏ̐

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Line( POINT2D *pPt, LINE2D *pLine )
{
	CIRCLE2D Cir = { pPt->Px, pPt->Py, 0.1f };

	return Circle_Line( &Cir, pLine );
}

//=============================================================================
/**
	_Ǝlp̓蔻s܂

	@param pPt		[in] ]Qcʏ̓_
	@param pRect	[in] ]Qcʏ̋`

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Rect( POINT2D *pPt, RECT2D *pRect )
{
	if ( pPt->Px < pRect->Sx ) return false;
	if ( pPt->Px > pRect->Ex ) return false;
	if ( pPt->Py < pRect->Sy ) return false;
	if ( pPt->Py > pRect->Ey ) return false;

	return true;
}

//=============================================================================
/**
	_Ɖ~̓蔻s܂

	@param pPt		[in] ]Qcʏ̓_
	@param pCir		[in] ]Qcʏ̉~

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Circle( POINT2D *pPt, CIRCLE2D *pCir )
{
	Float rx = pPt->Px - pCir->Center.Px;
	Float ry = pPt->Py - pCir->Center.Py;
	Float r = pCir->Radius;

	if ( ((rx * rx) + (ry * ry)) > (r * r) ) return false;

	return true;
}

//=============================================================================
/**
	_Ƒp`̓蔻s܂

	@param pPt		[in] ]Qcʏ̓_
	@param pPoly	[in] ]Qcʏ̃|S

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Polygon( POINT2D *pPt, POLYGON2D *pPoly )
{
	Sint32 num = 0;
	Sint32 cnt = pPoly->Count-1;

	for ( Sint32 i = 0; i < cnt; i++ )
	{
		Float x1 = pPoly->pPts[i+0].Px;
		Float y1 = pPoly->pPts[i+0].Py;
		Float x2 = pPoly->pPts[i+1].Px;
		Float y2 = pPoly->pPts[i+1].Py;

		Float x = pPt->Px;
		Float y = pPt->Py;

		Float n = (x - x1) * (y - y2) - (y - y1) * (x-x2);
		if ( n < 0 ) num++;
	}

	return ( (0 == num) || (num == cnt) );
}

//=============================================================================
/**
	lpƎlp̓蔻s܂

	@param pRect1	[in] ]Qcʏ̋`
	@param pRect2	[in] ]Qcʏ̋`

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Rect_Rect( RECT2D *pRect1, RECT2D *pRect2 )
{
	RECT2D Rect1 = *pRect1;
	RECT2D Rect2 = *pRect2;

	if ( Rect1.Sx > Rect1.Ex ) SWAP( Rect1.Sx, Rect1.Ex );
	if ( Rect1.Sy > Rect1.Ey ) SWAP( Rect1.Sy, Rect1.Ey );
	if ( Rect2.Sx > Rect2.Ex ) SWAP( Rect2.Sx, Rect2.Ex );
	if ( Rect2.Sy > Rect2.Ey ) SWAP( Rect2.Sy, Rect2.Ey );

	if ( Rect1.Sx > Rect2.Ex ) return false;
	if ( Rect1.Ex < Rect2.Sx ) return false;
	if ( Rect1.Sy > Rect2.Ey ) return false;
	if ( Rect1.Ey < Rect2.Sy ) return false;

	return true;
}

//=============================================================================
/**
	lpƉ~̓蔻s܂

	@param pRect	[in] ]Qcʏ̋`
	@param pCir		[in] ]Qcʏ̉~

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Rect_Circle( RECT2D *pRect, CIRCLE2D *pCir )
{
	POINT2D Pts[5] = {
		pRect->Sx, pRect->Sy,
		pRect->Ex, pRect->Sy,
		pRect->Ex, pRect->Ey,
		pRect->Sx, pRect->Ey,
		pRect->Sx, pRect->Sy,
	};

	POLYGON2D Poly = { 5, Pts };

	return Circle_Polygon( pCir, &Poly );
}

//=============================================================================
/**
	lpƒ̓蔻s܂

	@param pRect	[in] ]Qcʏ̋`
	@param pLine	[in] ]Qcʏ̐

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Rect_Line( RECT2D *pRect, LINE2D *pLine )
{
	LINE2D Lines[4] = {
		pRect->Sx, pRect->Sy, pRect->Ex, pRect->Sy,
		pRect->Ex, pRect->Sy, pRect->Ex, pRect->Ey,
		pRect->Ex, pRect->Ey, pRect->Sx, pRect->Ey,
		pRect->Sx, pRect->Ey, pRect->Sx, pRect->Sy,
	};

	for ( Sint32 i = 0; i < 4; i++ )
	{
		if ( Line_Line( &Lines[i], pLine ) )
		{
			return true;
		}
	}

	return false;
}

//=============================================================================
/**
	lpƑp`̓蔻s܂

	@param pRect	[in] ]Qcʏ̋`
	@param pPoly	[in] ]Qcʏ̃|S

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Rect_Polygon( RECT2D *pRect, POLYGON2D *pPoly )
{
	POINT2D Pts[5] = {
		pRect->Sx, pRect->Sy,
		pRect->Ex, pRect->Sy,
		pRect->Ex, pRect->Ey,
		pRect->Sx, pRect->Ey,
		pRect->Sx, pRect->Sy,
	};

	POLYGON2D Poly = {
		5, Pts,
	};

	return Polygon_Polygon( pPoly, &Poly );
}

//=============================================================================
/**
	~Ɖ~̓蔻s܂

	@param pCir1	[in] ]Qcʏ̉~
	@param pCir2	[in] ]Qcʏ̉~

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Circle_Circle( CIRCLE2D *pCir1, CIRCLE2D *pCir2 )
{
	Float rx = pCir1->Center.Px - pCir2->Center.Px;
	Float ry = pCir1->Center.Py - pCir2->Center.Py;
	Float r = pCir1->Radius + pCir2->Radius;

	return ( (rx * rx + ry * ry) < (r * r) );
}

//=============================================================================
/**
	~ƒ̓蔻s܂

	@param pCir		[in] ]Qcʏ̉~
	@param pLine	[in] ]Qcʏ̐

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Circle_Line( CIRCLE2D *pCir, LINE2D *pLine )
{
	//====================================================
	// ܂̗͒[~Ɋ܂܂΂nj
	//====================================================
	if ( Point_Circle( &pLine->Pt0, pCir ) ) return true;
	if ( Point_Circle( &pLine->Pt1, pCir ) ) return true;

	//====================================================
	// ~̒_ƒ܂ł̋߂.
	//====================================================

	// _璼ɐȒPʃxNg(t1)߂B
	static LVECTOR2 s1, s2, s3, t1, t2, v;
	static Float z;

	// xNg la - lb
	s1.x = pLine->Pt1.Px - pLine->Pt0.Px;
	s1.y = pLine->Pt1.Py - pLine->Pt0.Py;

	// xNg la - cir
	s2.x = pCir->Center.Px - pLine->Pt0.Px;
	s2.y = pCir->Center.Py - pLine->Pt0.Py;

	// xNg s1  s2 ̊O
	// QԂȂ̂łyɕsȃxNg
	z = s1.CrossProduct( &s2 );

	// ~̒Sɑ݂ƂiOςOɂȂ
	if ( z == 0 )
	{
		v.x = pCir->Center.Px;
		v.y = pCir->Center.Py;
	}
	else
	{
		// xNg s1  s3 ̊O
		// ̃xNg la - lb ̐ɂȂ
		s3.x = +(s1.y * z);
		s3.y = -(s1.x * z);

		// xNg̐K
		t1.Normalize( &s3 );

		// ς狗Zo
		t2.x = pLine->Pt0.Px - pCir->Center.Px;
		t2.y = pLine->Pt0.Py - pCir->Center.Py;
		Float len = t1.DotProduct( &t2 );

		// _Ɖ~̒S̋aȏȂmf
		if ( len > pCir->Radius ) return false;

		// ~̒S la - lb Ɉ̑̍W߂
		v.x = pCir->Center.Px + (t1.x * len);
		v.y = pCir->Center.Py + (t1.y * len);
	}

	// v  pLine->Pt0 - pLine->Pt1 ̊Ԃɂ΂nj
	if ( (pLine->Pt0.Px < v.x) && (v.x < pLine->Pt1.Px) ) return true;
	if ( (pLine->Pt1.Px < v.x) && (v.x < pLine->Pt0.Px) ) return true;
	if ( (pLine->Pt0.Py < v.y) && (v.y < pLine->Pt1.Py) ) return true;
	if ( (pLine->Pt1.Py < v.y) && (v.y < pLine->Pt0.Py) ) return true;

	return false;
}

//=============================================================================
/**
	~Ƒp`̓蔻s܂

	@param pCir		[in] ]Qcʏ̉~
	@param pPoly	[in] ]Qcʏ̃|S

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Circle_Polygon( CIRCLE2D *pCir, POLYGON2D *pPoly )
{
	Sint32 cnt = pPoly->Count - 1;

	// ~̒Sp`Ɋ܂܂ĂȂnj
	if ( Point_Polygon( &pCir->Center, pPoly ) ) return true;

	// p`̊eӂƉ~̌
	for ( Sint32 i = 0; i < cnt; i++ )
	{
		LINE2D Line = {
			pPoly->pPts[i+0].Px, pPoly->pPts[i+0].Py,
			pPoly->pPts[i+1].Px, pPoly->pPts[i+1].Py
		};

		if ( Circle_Line( pCir, &Line ) ) return true;
	}

	return false;
}

//=============================================================================
/**
	ƒ̓蔻s܂

	@param pLine1	[in] ]Qcʏ̐
	@param pLine2	[in] ]Qcʏ̐

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Line_Line( LINE2D *pLine1, LINE2D *pLine2 )
{
	static Float n1, n2;

	//-------------------------------------------
	// l1  l2 ܂ł邩ۂ
	//-------------------------------------------
	n1 = Side( pLine2->Pt0.Px, pLine2->Pt0.Py, pLine1 );
	n2 = Side( pLine2->Pt1.Px, pLine2->Pt1.Py, pLine1 );
	Float ret1 = (n1 * n2);

	//-------------------------------------------
	// l2  l1 ܂ł邩ۂ
	//-------------------------------------------
	n1 = Side( pLine1->Pt0.Px, pLine1->Pt0.Py, pLine2 );
	n2 = Side( pLine1->Pt1.Px, pLine1->Pt1.Py, pLine2 );
	Float ret2 = (n1 * n2);

	return ((ret1<=0) && (ret2<=0) );
}

//=============================================================================
/**
	Ƒp`̓蔻s܂

	@param pLine	[in] ]Qcʏ̐
	@param pPoly	[in] ]Qcʏ̃|S

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Line_Polygon( LINE2D *pLine, POLYGON2D *pPoly )
{
	Sint32 cnt = pPoly->Count;

	// ̊e[p`Ɋ܂܂Ă邩`FbN
	if ( Point_Polygon( &pLine->Pt0, pPoly ) ) return true;
	if ( Point_Polygon( &pLine->Pt1, pPoly ) ) return true;

	// p`̊eӂƒ̌
	for ( Sint32 i = 0; i < cnt - 1; i++ )
	{
		LINE2D Line = {
			pPoly->pPts[i+0].Px, pPoly->pPts[i+0].Py,
			pPoly->pPts[i+1].Px, pPoly->pPts[i+1].Py
		};

		if ( Line_Line( &Line, pLine ) ) return true;
	}
	return false;
}

//=============================================================================
/**
	p`Ƒp`̓蔻s܂

	@param pPoly1	[in] ]Qcʏ̃|S
	@param pPoly2	[in] ]Qcʏ̃|S

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Polygon_Polygon( POLYGON2D *pPoly1, POLYGON2D *pPoly2 )
{
	Sint32 cnt1 = pPoly1->Count-1;
	Sint32 cnt2 = pPoly2->Count-1;

	//==========================================================
	// oEfBO{bNXZo
	//==========================================================
	Float x11 = pPoly1->pPts[0].Px, x12 = pPoly1->pPts[0].Px;
	Float y11 = pPoly1->pPts[0].Py, y12 = pPoly1->pPts[0].Py;
	Float x21 = pPoly2->pPts[0].Px, x22 = pPoly2->pPts[0].Px;
	Float y21 = pPoly2->pPts[0].Py, y22 = pPoly2->pPts[0].Py;

	for ( Sint32 i = 1; i < cnt1; i++ )
	{
		if ( pPoly1->pPts[i].Px < x11 ) x11 = pPoly1->pPts[i].Px;
		ef ( pPoly1->pPts[i].Px > x12 ) x12 = pPoly1->pPts[i].Px;
		if ( pPoly1->pPts[i].Py < y11 ) y11 = pPoly1->pPts[i].Py;
		ef ( pPoly1->pPts[i].Py > y12 ) y12 = pPoly1->pPts[i].Py;
	}
	for ( Sint32 i = 1; i < cnt2; i++ )
	{
		if ( pPoly2->pPts[i].Px < x21 ) x21 = pPoly2->pPts[i].Px;
		ef ( pPoly2->pPts[i].Px > x22 ) x22 = pPoly2->pPts[i].Px;
		if ( pPoly2->pPts[i].Py < y21 ) y21 = pPoly2->pPts[i].Py;
		ef ( pPoly2->pPts[i].Py > y22 ) y22 = pPoly2->pPts[i].Py;
	}

	//----------------------------------------------------------
	// oEfBO{bNX̐ڐG
	//----------------------------------------------------------
	if ( x11 > x22 ) return false;
	if ( x12 < x21 ) return false;
	if ( y11 > y22 ) return false;
	if ( y12 < y21 ) return false;

	//==========================================================
	// OςŔ
	//==========================================================

	//----------------------------------------------------------
	// p1̊e_p2̋`ɓĂ邩
	//----------------------------------------------------------
	for ( Sint32 i = 0; i < cnt1; i++ )
	{
		if ( Point_Polygon( &pPoly1->pPts[i], pPoly2 ) )
		{
			return true;
		}
	}

	//----------------------------------------------------------
	// p2̊e_p1̋`ɓĂ邩
	//----------------------------------------------------------
	for ( Sint32 i = 0; i < cnt2; i++ )
	{
		if ( Point_Polygon( &pPoly2->pPts[i], pPoly1 ) )
		{
			return true;
		}
	}

	//==========================================================
	// e_̌
	//==========================================================
	for ( Sint32 i = 0; i < cnt1; i++ )
	{
		for ( Sint32 j = 0; j < cnt2; j++ )
		{
			LINE2D Line1 = {
				pPoly1->pPts[i+0].Px, pPoly1->pPts[i+0].Py,
				pPoly1->pPts[i+1].Px, pPoly1->pPts[i+1].Py
			};

			LINE2D Line2 = {
				pPoly2->pPts[j+0].Px, pPoly2->pPts[j+0].Py,
				pPoly2->pPts[j+1].Px, pPoly2->pPts[j+1].Py
			};

			if ( Line_Line( &Line1, &Line2 ) ) return true;
		}
	}

	return false;
}


//===================================================================================
//===================================================================================
//
//
//			Rc֘A
//
//
//===================================================================================
//===================================================================================


//=============================================================================
/**
	_Ɠ_̓蔻s܂

	@param pPt1	[in] ]Rcԓ̓_
	@param pPt2	[in] ]Rcԓ̓_

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Point_3D( POINT3D *pPt1, POINT3D *pPt2 )
{
	if ( pPt1->Px != pPt2->Px ) return false;
	if ( pPt1->Py != pPt2->Py ) return false;
	if ( pPt1->Pz != pPt2->Pz ) return false;

	return true;
}

//=============================================================================
/**
	_Ƌ̓蔻s܂

	@param pPt		[in] ]Rcԓ̓_
	@param pSph		[in] ]Rcԓ̋

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Sphere_3D( POINT3D *pPt, SPHERE3D *pSph )
{
	SPHERE3D Cir( *pPt, 0.0f );

	return Sphere_Sphere_3D( pSph, &Cir );
}

//=============================================================================
/**
	_ƒ̓蔻s܂

	@param pPt		[in] ]Rcԓ̓_
	@param pLine	[in] ]Rcԓ̐

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Line_3D( POINT3D *pPt, LINE3D *pLine )
{
	SPHERE3D Cir( *pPt, 1.0f );

	return Sphere_Line_3D( &Cir, pLine );
}

//=============================================================================
/**
	_Ɩʂ̓蔻s܂

	@param pPt		[in] ]Rcԓ̓_
	@param pPlane	[in] ]Rcԓ̕

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Plane_3D( POINT3D *pPt, PLANE3D *pPlane )
{
	return false;
}

//=============================================================================
/**
	_Ɣ̓蔻s܂

	@param pPt	[in] ]Rcԓ̓_
	@param pBox	[in] ]Rcԓ̔

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Point_Box_3D( POINT3D *pPt, BOX3D *pBox )
{
	Uint8 bOutside = 0x00;
	for ( Sint32 i = 0; i < 6; i++ )
	{
		Float res =
			pBox->Planes[i].a * pPt->Px +
			pBox->Planes[i].b * pPt->Py +
			pBox->Planes[i].c * pPt->Pz +
			pBox->Planes[i].d;

		if ( res < 0 ) bOutside |= (1 << i);
	}

	// |Cg6̕ʂׂĂ̓łꍇ{bNXɂ܂
	return (bOutside == 0);
}

//=============================================================================
/**
	Ƌ̓蔻s܂

	@param pSph1	[in] ]Rcԓ̋
	@param pSph2	[in] ]Rcԓ̋

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Sphere_Sphere_3D( SPHERE3D *pSph1, SPHERE3D *pSph2 )
{
	Float rx = pSph1->Center.Px + pSph2->Center.Px;
	Float ry = pSph1->Center.Py + pSph2->Center.Py;
	Float rz = pSph1->Center.Pz + pSph2->Center.Pz;
	Float r = pSph1->Radius + pSph2->Radius;

	return ( (rx * rx) + (ry * ry) + (rz * rz) < (r * r) );
}

//=============================================================================
/**
	ƒ̓蔻s܂

	@param pSph		[in] ]Rcԓ̋
	@param pLine	[in] ]Rcԓ̐

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Sphere_Line_3D( SPHERE3D *pSph, LINE3D *pLine )
{
	//====================================================
	// ܂̗͒[Ɋ܂܂Ă邩`FbN
	//====================================================
	Float r = pSph->Radius;
	Float rx1 = pSph->Center.Px - pLine->Pt0.Px;
	Float ry1 = pSph->Center.Py - pLine->Pt0.Py;
	Float rz1 = pSph->Center.Pz - pLine->Pt0.Pz;
	Float rx2 = pSph->Center.Px - pLine->Pt1.Px;
	Float ry2 = pSph->Center.Py - pLine->Pt1.Py;
	Float rz2 = pSph->Center.Pz - pLine->Pt1.Pz;

	if ( ((rx1 * rx1) + (ry1 * ry1) + (rz1 * rz1)) < (r * r) ) return true;
	if ( ((rx2 * rx2) + (ry2 * ry2) + (rz2 * rz2)) < (r * r) ) return true;

	//====================================================
	// ̒_ƒ܂ł̋߂.
	//====================================================

	// _璼ɐȒPʃxNg(t1)߂B
	static LVECTOR3 s1, s2, s3, s4, t1, t2, v;

	// xNg la - lb
	s1.x = pLine->Pt1.Px - pLine->Pt0.Px;
	s1.y = pLine->Pt1.Py - pLine->Pt0.Py;
	s1.z = pLine->Pt1.Pz - pLine->Pt0.Pz;

	// xNg la - cir
	s2.x = pSph->Center.Px - pLine->Pt0.Px;
	s2.y = pSph->Center.Py - pLine->Pt0.Py;
	s2.z = pSph->Center.Pz - pLine->Pt0.Pz;

	// s1, s2 ̊O s3
	s3.CrossProduct( &s1, &s2 );

	// s1, s3 ̊O s4 = 
	s4.CrossProduct( &s1, &s3 );

	// K
	t1.Normalize( &s4 );

	// ς狗Zo
	t2.x = pLine->Pt0.Px - pSph->Center.Px;
	t2.y = pLine->Pt0.Py - pSph->Center.Py;
	t2.z = pLine->Pt0.Pz - pSph->Center.Pz;
	Float len = t1.DotProduct( &t2 );

	// _Ɖ~̒S̋aȏȂmf
	if ( len > r ) return false;

	// ̒S(cir)璼(pLine->a, pLine->b)Ɉ̑̍W(v)
	v.x = pSph->Center.Px + (t1.x * len);
	v.y = pSph->Center.Py + (t1.y * len);
	v.z = pSph->Center.Pz + (t1.z * len);
	
	// vpLine->apLine->b̊Ԃɂ΂nj
	if ( (pLine->Pt0.Px < v.x) && (v.x < pLine->Pt1.Px) ) return true;
	if ( (pLine->Pt1.Px < v.x) && (v.x < pLine->Pt0.Px) ) return true;
	if ( (pLine->Pt0.Py < v.y) && (v.y < pLine->Pt1.Py) ) return true;
	if ( (pLine->Pt1.Py < v.y) && (v.y < pLine->Pt0.Py) ) return true;
	if ( (pLine->Pt0.Pz < v.z) && (v.z < pLine->Pt1.Pz) ) return true;
	if ( (pLine->Pt1.Pz < v.z) && (v.z < pLine->Pt0.Pz) ) return true;

	return false;
}

//=============================================================================
/**
	Ɩʂ̓蔻s܂<BR>
	̊֐͖ł̂ŁA̕@ł̔𐄏܂B

	@param pSph		[in] ]Rcԓ̋
	@param pPlane	[in] ]Rcԓ̕

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Sphere_Plane( SPHERE3D *pSph, PLANE3D *pPlane )
{
	return false;
}

//=============================================================================
/**
	Ɣ̓蔻s܂<BR>
	̊֐͖ł̂ŁA̕@ł̔𐄏܂B

	@param pSph		[in] ]Rcԓ̋
	@param pBox		[in] ]Rcԓ̔

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Sphere_Box_3D( SPHERE3D *pSph, BOX3D *pBox )
{
	// ̒S̒ɓĂ΂nj
	if ( Point_Box_3D( &pSph->Center, pBox ) ) return true;

	return false;
}

//=============================================================================
/**
	Ɩʂ̓蔻s܂

	@param pLine	[in] ]Rcԓ̐
	@param pPlane	[in] ]Rcԓ̕
	@param pPoints	[in] pPlane̔Cӂ̂R_
	@param pPos		[out] ՓˍWi[


	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Line_Plane_3D( LINE3D *pLine, PLANE3D *pPlane, POINT3D *pPoints, POINT3D *pPos )
{
	LVECTOR3 VecP;
	LVECTOR3 VecQ;
	LVECTOR3 VecT;
	Float fDet;
	Float fInvDet;
	Float ParamT, ParamU, ParamV;

	// ̗[ʂ܂ő݂Ă邩`FbN
	Float fDist0 = (pPlane->a * pLine->Pt0.Px) + (pPlane->b * pLine->Pt0.Py) + (pPlane->c * pLine->Pt0.Pz) + pPlane->d;
	Float fDist1 = (pPlane->a * pLine->Pt1.Px) + (pPlane->b * pLine->Pt1.Py) + (pPlane->c * pLine->Pt1.Pz) + pPlane->d;
	if ( (fDist0 * fDist1) >= 0.0f ) return false;

	LVECTOR3 Start = LVECTOR3(pLine->Pt0.Px,pLine->Pt0.Py,pLine->Pt0.Pz);
	LVECTOR3 Dir = LVECTOR3(pLine->Pt1.Px,pLine->Pt1.Py,pLine->Pt1.Pz) - Start;
	LVECTOR3 Vx[] = {
		LVECTOR3(pPoints[0].Px,pPoints[0].Py,pPoints[0].Pz),
		LVECTOR3(pPoints[1].Px,pPoints[1].Py,pPoints[1].Pz),
		LVECTOR3(pPoints[2].Px,pPoints[2].Py,pPoints[2].Pz),
	};

	// Vx[0]LQxNgZo
	LVECTOR3 Edge1 = Vx[1] - Vx[0];
	LVECTOR3 Edge2 = Vx[2] - Vx[0];

	// Up[^[vZ
	VecP.CrossProduct( &Dir, &Edge2 );

    // ςOɋ߂ꍇ͎Op`̕ʂɂ
	fDet = Edge1.DotProduct( &VecP );
	if ( fDet > 0 )
	{
		VecT = Start - Vx[0];
	}
	else
	{
		VecT = Vx[0] - Start;
		fDet = -fDet;
	}

	if ( fDet < 0.0001f ) return false;

	// Up[^[уeXgEvZ
	ParamU = VecT.DotProduct( &VecP );
	if ( (ParamU < 0.0f) || (ParamU > fDet) ) return false;

	// Vp[^[eXg鏀܂
	VecQ.CrossProduct( &VecT, &Edge1 );

	// Vp[^[уeXgEvZ
	ParamV = Dir.DotProduct( &VecQ );
	if ( (ParamV < 0.0f) || ((ParamU + ParamV) > fDet) ) return false;

	// TvZAXP[O
	ParamT = Edge2.DotProduct( &VecQ );
	fInvDet = 1.0f / fDet;
	ParamT *= fInvDet;
	ParamU *= fInvDet;
	ParamV *= fInvDet;

	// ʒuۑ
	if ( pPos != NULL )
	{
		pPos->Px = Vx[0].x + (ParamU * (Vx[1].x - Vx[0].x)) + (ParamV * (Vx[2].x - Vx[0].x));
		pPos->Py = Vx[0].y + (ParamU * (Vx[1].y - Vx[0].y)) + (ParamV * (Vx[2].y - Vx[0].y));
		pPos->Pz = Vx[0].z + (ParamU * (Vx[1].z - Vx[0].z)) + (ParamV * (Vx[2].z - Vx[0].z));
	}

	return true;
}

//=============================================================================
/**
	Ɣ̓蔻s܂

	@param pLine	[in] ]Rcԓ̐
	@param pBox		[in] ]Rcԓ̔

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Line_Box_3D( LINE3D *pLine, BOX3D *pBox )
{
	// ̒ɓ_Ă邩`FbN
	if ( Point_Box_3D( &pLine->Pt0, pBox ) ) return true;
	if ( Point_Box_3D( &pLine->Pt1, pBox ) ) return true;

	// ʂƐĂ邩`FbN
	POINT3D Face[][4] =
	{
		{ pBox->Points[0], pBox->Points[2], pBox->Points[3], pBox->Points[1] }, // front
		{ pBox->Points[4], pBox->Points[5], pBox->Points[7], pBox->Points[6] }, // back
		{ pBox->Points[0], pBox->Points[4], pBox->Points[6], pBox->Points[2] }, // left
		{ pBox->Points[1], pBox->Points[3], pBox->Points[7], pBox->Points[5] }, // right
		{ pBox->Points[2], pBox->Points[6], pBox->Points[7], pBox->Points[3] }, // top
		{ pBox->Points[0], pBox->Points[4], pBox->Points[5], pBox->Points[1] }, // bottom
	};

	for ( Sint32 j = 0; j < 6; j++ )
	{
		if ( EdgeIntersectsQuad( (POINT3D*)pLine, Face[j], &pBox->Planes[j], NULL ) )
		{
			return true;
		}
	}

	//=========================================================================================================

	return false;
}

//=============================================================================
/**
	ƎOp`̓蔻s܂

	@param pLine	[in] ]Rcԓ̐
	@param Points	[in] ]Rcԓ̎Op`\_
	@param pT		[out] Tp[^[i[
	@param pU		[out] Up[^[i[
	@param pV		[out] Vp[^[i[

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Line_Triangle_3D( LINE3D* pLine, POINT3D Points[], Float *pT, Float *pU, Float *pV )
{
	LVECTOR3 VecP;
	LVECTOR3 VecQ;
	LVECTOR3 VecT;
	Float fDet;
	Float fInvDet;

	LVECTOR3 Start = LVECTOR3(pLine->Pt0.Px,pLine->Pt0.Py,pLine->Pt0.Pz);
	LVECTOR3 Dir = LVECTOR3(pLine->Pt1.Px,pLine->Pt1.Py,pLine->Pt1.Pz) - Start;
	LVECTOR3 Vx[] = {
		LVECTOR3(Points[0].Px,Points[0].Py,Points[0].Pz),
		LVECTOR3(Points[1].Px,Points[1].Py,Points[1].Pz),
		LVECTOR3(Points[2].Px,Points[2].Py,Points[2].Pz),
	};

	Dir.Normalize();

	// pV0LQxNgZo
	LVECTOR3 Edge1 = Vx[1] - Vx[0];
	LVECTOR3 Edge2 = Vx[2] - Vx[0];

	// Up[^[vZ
	VecP.CrossProduct( &Dir, &Edge2 );

    // ςOɋ߂ꍇ͎Op`̕ʂɂ
	fDet = Edge1.DotProduct( &VecP );
	if ( fDet > 0 )
	{
		VecT = Start - Vx[0];
	}
	else
	{
		VecT = Vx[0] - Start;
		fDet = -fDet;
	}

	if ( fDet < 0.0001f ) return false;

	// Up[^[уeXgEvZ
	*pU = VecT.DotProduct( &VecP );
	if ( (*pU < 0.0f) || (*pU > fDet) ) return false;

	// Vp[^[eXg鏀܂
	VecQ.CrossProduct( &VecT, &Edge1 );

	// Vp[^[уeXgEvZ
	*pV = Dir.DotProduct( &VecQ );
	if ( (*pV < 0.0f) || ((*pU + *pV) > fDet) ) return false;

	// TvZAXP[O
	*pT = Edge2.DotProduct( &VecQ );
	fInvDet = 1.0f / fDet;
	*pT *= fInvDet;
	*pU *= fInvDet;
	*pV *= fInvDet;

	return true;
}

//=============================================================================
/**
	ʂƖʂ̐ڐG܂

	@param pPlane1			[in] ʂP
	@param pPlane2			[in] ʂQ
	@param Points1[]		[in] ʂP\S_
	@param Points2[]		[in] ʂQ\S_

	@retval true		Ă
	@retval false		ĂȂ
*/
//=============================================================================
bool LunaCollision::Plane_Plane_3D( PLANE3D *pPlane1, PLANE3D *pPlane2, POINT3D Points1[], POINT3D Points2[] )
{
	POINT3D Edge[2];

	Edge[0] = Points1[0]; Edge[1] = Points1[1];
	if ( EdgeIntersectsQuad( Edge, Points2, pPlane2, NULL ) ) return true;

	Edge[0] = Points1[1]; Edge[1] = Points1[2];
	if ( EdgeIntersectsQuad( Edge, Points2, pPlane2, NULL ) ) return true;

	Edge[0] = Points1[2]; Edge[1] = Points1[3];
	if ( EdgeIntersectsQuad( Edge, Points2, pPlane2, NULL ) ) return true;

	Edge[0] = Points1[3]; Edge[1] = Points1[4];
	if ( EdgeIntersectsQuad( Edge, Points2, pPlane2, NULL ) ) return true;

	Edge[0] = Points2[0]; Edge[1] = Points2[1];
	if ( EdgeIntersectsQuad( Edge, Points1, pPlane1, NULL ) ) return true;

	Edge[0] = Points2[1]; Edge[1] = Points2[2];
	if ( EdgeIntersectsQuad( Edge, Points1, pPlane1, NULL ) ) return true;

	Edge[0] = Points2[2]; Edge[1] = Points2[3];
	if ( EdgeIntersectsQuad( Edge, Points1, pPlane1, NULL ) ) return true;

	Edge[0] = Points2[3]; Edge[1] = Points2[4];
	if ( EdgeIntersectsQuad( Edge, Points1, pPlane1, NULL ) ) return true;

	return false;
}

//=============================================================================
/**
	ʂƒ̌_Zo܂

	@param Intersection		[out] _i[
	@param pPlane			[in] 
	@param pEdge0			[in] ̒[
	@param pEdge1			[in] ̒[

	@retval true		Ă
	@retval false		ĂȂ
*/
//=============================================================================
bool LunaCollision::PlaneIntersectLine( POINT3D *Intersection, PLANE3D *pPlane, POINT3D *pEdge0, POINT3D *pEdge1 )
{
	// xNg
	POINT3D Dir( pEdge1->Px - pEdge0->Px, pEdge1->Py - pEdge0->Py, pEdge1->Pz - pEdge0->Pz );

	// ̓_̕ʂ̋_ʂŗɂ邩ǂ𒲂ׂ
	Float a0 = (pPlane->a * pEdge0->Px) + (pPlane->b * pEdge0->Py) + (pPlane->c * pEdge0->Pz) + pPlane->d;
	Float a1 = (pPlane->a * pEdge1->Px) + (pPlane->b * pEdge1->Py) + (pPlane->c * pEdge1->Pz) + pPlane->d;
	if ( (a0 < 0.0f) && (a1 < 0.0f) ) return false;
	if ( (a0 > 0.0f) && (a1 > 0.0f) ) return false;

	// }ϐ߂
	Float t = -((pPlane->a * pEdge0->Px) + (pPlane->b * pEdge0->Py) + (pPlane->c * pEdge0->Pz) + pPlane->d) / ((pPlane->a * Dir.Px) + (pPlane->b * Dir.Py) + (pPlane->c * Dir.Pz));
	if ( (t < 0.0f) || (1.0f < t) ) return false;

	// }ϐ_߂
	Intersection->Px = pEdge0->Px + (Dir.Px * t);
	Intersection->Py = pEdge0->Py + (Dir.Py * t);
	Intersection->Pz = pEdge0->Pz + (Dir.Pz * t);

	return true;
}

//=============================================================================
/**
	S_ō\镽ʂƒ̌s܂

	@param pEdges		[in] 
	@param pFacePoints	[in] ʂ̓_iS_j
	@param pPlane		[in] `FbN镽
	@param pHitPos		[out] ʂƂ̌ʒu

	@retval true		Ă
	@retval false		ĂȂ
*/
//=============================================================================
bool LunaCollision::EdgeIntersectsQuad( POINT3D* pEdges, POINT3D* pFacePoints, PLANE3D* pPlane, POINT3D *pHitPos )
{
	POINT3D Intersection;
	POINT2D FacePoints[5];
	POINT2D Point;

	// [̃|Cgʂ̓ɂꍇAʂƌ܂
	Float fDist1 = pPlane->a * pEdges[0].Px + pPlane->b * pEdges[0].Py + pPlane->c * pEdges[0].Pz + pPlane->d;
	Float fDist2 = pPlane->a * pEdges[1].Px + pPlane->b * pEdges[1].Py + pPlane->c * pEdges[1].Pz + pPlane->d;
	if ( (fDist1 * fDist2) > 0 ) return false;

	// ѕʂ̊Ԃ̌_܂
	//  炪s̏ꍇAƖʂ͌܂
	//  ̂Ƃ ::D3DXPlaneIntersectLine()  NULL Ԃ܂
	if ( !PlaneIntersectLine( &Intersection, pPlane, &pEdges[0], &pEdges[1] ) ) return false;
	if ( pHitPos != NULL ) *pHitPos = Intersection;

	// |S̒_ʒueXg邽߂Qcʏւ̃vWFNg
	Float fAbsA = (pPlane->a > 0 ? pPlane->a : -pPlane->a);
	Float fAbsB = (pPlane->b > 0 ? pPlane->b : -pPlane->b);
	Float fAbsC = (pPlane->c > 0 ? pPlane->c : -pPlane->c);
	if ( (fAbsA > fAbsB) && (fAbsA > fAbsC) )
	{
		// wsȂ̂łxƂygpČvZ
		for( Sint32 i = 0; i < 4; i++)
		{
			FacePoints[i].Px = pFacePoints[i].Py;
			FacePoints[i].Py = pFacePoints[i].Pz;
		}
		Point.Px = Intersection.Py;
		Point.Py = Intersection.Pz;
	}
	ef ( (fAbsB > fAbsA) && (fAbsB > fAbsC) )
	{
		// xsȂ̂łwƂygpČvZ
		for ( Sint32 i = 0; i < 4; i++)
		{
			FacePoints[i].Px = pFacePoints[i].Px;
			FacePoints[i].Py = pFacePoints[i].Pz;
		}
		Point.Px = Intersection.Px;
		Point.Py = Intersection.Pz;
	}
	else
	{
		// ysȂ̂łwƂxgpČvZ
		for ( Sint32 i = 0; i < 4; i++)
		{
			FacePoints[i].Px = pFacePoints[i].Px;
			FacePoints[i].Py = pFacePoints[i].Py;
		}
		Point.Px = Intersection.Px;
		Point.Py = Intersection.Py;
	}

	FacePoints[4] = FacePoints[0];

	// _ʂ̂̊OɂꍇA̒_ʂ̊Oɂ܂
	// ̃}gbNX̌vfƂ邱Ƃɂ肱邱Ƃł܂
	// | x0 y0 1 |
	// | x1 y1 1 |
	// | x2 y2 1 |
	Float x0 = FacePoints[0].Px;
	Float y0 = FacePoints[0].Py;
	Float x1 = FacePoints[1].Px;
	Float y1 = FacePoints[1].Py;
	Float x2 = FacePoints[2].Px;
	Float y2 = FacePoints[2].Py;
	bool bClockwise = ((x1*y2 - y1*x2 - x0*y2 + y0*x2 + x0*y1 - y0*x1) < 0);

	x2 = Point.Px;
	y2 = Point.Py;
	for ( Sint32 i = 0; i < 4; i++ )
	{
		x0 = FacePoints[i].Px;
		y0 = FacePoints[i].Py;
		x1 = FacePoints[i+1].Px;
		y1 = FacePoints[i+1].Py;

		if ( ((x1*y2 - y1*x2 - x0*y2 + y0*x2 + x0*y1 - y0*x1) > 0) == bClockwise ) return false;
	}

	// |Cg͖ʂׂĂ̓ɂ܂
	return true;
}

//=============================================================================
/**
	Ɣ̓蔻s܂

	@param pBox1	[in] ]Rcԓ̔
	@param pBox2	[in] ]Rcԓ̔

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::Box_Box_3D( BOX3D *pBox1, BOX3D *pBox2 )
{
	Uint8 bOutside[8];

	//=========================================================================================================
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	// ݂̔̒Ɍ݂̓_Ă邩`FbN
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	//=========================================================================================================
	for ( Sint32 i = 0; i < 8; i++ )
	{
		bOutside[i] = 0x00;
		for ( Sint32 j = 0; j < 6; j++ )
		{
			Float fRet =
				pBox1->Planes[j].a * pBox2->Points[i].Px +
				pBox1->Planes[j].b * pBox2->Points[i].Py +
				pBox1->Planes[j].c * pBox2->Points[i].Pz +
				pBox1->Planes[j].d;
			
			if ( fRet < 0 ) bOutside[i] |= (1 << j);
		}

		// |CgU̃r[tX^ʂׂĂ̓łꍇ
		// r[tX^ɂ܂
		if ( bOutside[i] == 0 ) return true;
	}

	// |CgׂĔCӂ̒P̃r[tX^ʂ̊Oɂꍇ
	// IuWFNg̓r[tX^Oɂ܂B
	if ( (bOutside[0] & bOutside[1] & bOutside[2] & bOutside[3] & bOutside[4] & bOutside[5] & bOutside[6] & bOutside[7]) != 0x00 )
	{
		for ( Sint32 i = 0; i < 8; i++ )
		{
			bOutside[i] = 0x00;
			for ( Sint32 j = 0; j < 6; j++ )
			{
				Float fRet =
					pBox2->Planes[j].a * pBox1->Points[i].Px +
					pBox2->Planes[j].b * pBox1->Points[i].Py +
					pBox2->Planes[j].c * pBox1->Points[i].Pz +
					pBox2->Planes[j].d;
				
				if ( fRet < 0 ) bOutside[i] |= (1 << j);
			}

			// |CgU̃r[tX^ʂׂĂ̓łꍇ
			// r[tX^ɂ܂
			if ( bOutside[i] == 0 ) return true;
		}

		// |CgׂĔCӂ̒P̃r[tX^ʂ̊Oɂꍇ
		// IuWFNg̓r[tX^Oɂ܂B
		if ( (bOutside[0] & bOutside[1] & bOutside[2] & bOutside[3] & bOutside[4] & bOutside[5] & bOutside[6] & bOutside[7]) != 0x00 )
		{
			return false;
		}
	}

	//=========================================================================================================
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	//=========================================================================================================
	POINT3D Edge1[][2] = 
	{
		{ pBox2->Points[0], pBox2->Points[1] }, // front bottom
		{ pBox2->Points[2], pBox2->Points[3] }, // front top
		{ pBox2->Points[0], pBox2->Points[2] }, // front left
		{ pBox2->Points[1], pBox2->Points[3] }, // front right
		{ pBox2->Points[4], pBox2->Points[5] }, // back bottom
		{ pBox2->Points[6], pBox2->Points[7] }, // back top
		{ pBox2->Points[4], pBox2->Points[6] }, // back left
		{ pBox2->Points[5], pBox2->Points[7] }, // back right
		{ pBox2->Points[0], pBox2->Points[4] }, // left bottom
		{ pBox2->Points[2], pBox2->Points[6] }, // left top
		{ pBox2->Points[1], pBox2->Points[5] }, // right bottom
		{ pBox2->Points[3], pBox2->Points[7] }, // right top
	};
	POINT3D Face1[][4] =
	{
		{ pBox1->Points[0], pBox1->Points[2], pBox1->Points[3], pBox1->Points[1] }, // front
		{ pBox1->Points[4], pBox1->Points[5], pBox1->Points[7], pBox1->Points[6] }, // back
		{ pBox1->Points[0], pBox1->Points[4], pBox1->Points[6], pBox1->Points[2] }, // left
		{ pBox1->Points[1], pBox1->Points[3], pBox1->Points[7], pBox1->Points[5] }, // right
		{ pBox1->Points[2], pBox1->Points[6], pBox1->Points[7], pBox1->Points[3] }, // top
		{ pBox1->Points[0], pBox1->Points[4], pBox1->Points[5], pBox1->Points[1] }, // bottom
	};

	for( Sint32 i = 0; i < 12; i++ )
	{
		for( Sint32 j = 0; j < 6; j++ )
		{
			if ( EdgeIntersectsQuad( Edge1[i], Face1[j], &pBox1->Planes[j], NULL ) )
			{
				return true;
			}
		}
	}

	//=========================================================================================================
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	//=========================================================================================================
	POINT3D Edge2[][2] = 
	{
		{ pBox1->Points[0], pBox1->Points[1] }, // front bottom
		{ pBox1->Points[2], pBox1->Points[3] }, // front top
		{ pBox1->Points[0], pBox1->Points[2] }, // front left
		{ pBox1->Points[1], pBox1->Points[3] }, // front right
		{ pBox1->Points[4], pBox1->Points[5] }, // back bottom
		{ pBox1->Points[6], pBox1->Points[7] }, // back top
		{ pBox1->Points[4], pBox1->Points[6] }, // back left
		{ pBox1->Points[5], pBox1->Points[7] }, // back right
		{ pBox1->Points[0], pBox1->Points[4] }, // left bottom
		{ pBox1->Points[2], pBox1->Points[6] }, // left top
		{ pBox1->Points[1], pBox1->Points[5] }, // right bottom
		{ pBox1->Points[3], pBox1->Points[7] }, // right top
	};
	POINT3D Face2[][4] =
	{
		{ pBox2->Points[0], pBox2->Points[2], pBox2->Points[3], pBox2->Points[1] }, // front
		{ pBox2->Points[4], pBox2->Points[5], pBox2->Points[7], pBox2->Points[6] }, // back
		{ pBox2->Points[0], pBox2->Points[4], pBox2->Points[6], pBox2->Points[2] }, // left
		{ pBox2->Points[1], pBox2->Points[3], pBox2->Points[7], pBox2->Points[5] }, // right
		{ pBox2->Points[2], pBox2->Points[6], pBox2->Points[7], pBox2->Points[3] }, // top
		{ pBox2->Points[0], pBox2->Points[4], pBox2->Points[5], pBox2->Points[1] }, // bottom
	};

	for( Sint32 i = 0; i < 12; i++ )
	{
		for( Sint32 j = 0; j < 6; j++ )
		{
			if ( EdgeIntersectsQuad( Edge2[i], Face2[j], &pBox2->Planes[j], NULL ) )
			{
				return true;
			}
		}
	}

	//=========================================================================================================
	// r[tX^oEfBO{bNXɊ܂܂邩ۂ
	// r[tX^̑S_oEfBO{bNX̊Oɂꍇ
	// r[tX^̓oEfBO{bNXɂ͊܂܂܂
	//=========================================================================================================
	for ( Sint32 i = 0; i < 6; i++ )
	{
		for ( Sint32 j = 0; j < 8; j++ )
		{
			Float fRet =
				pBox1->Planes[i].a * pBox2->Points[j].Px +
				pBox1->Planes[i].b * pBox2->Points[j].Py +
				pBox1->Planes[i].c * pBox2->Points[j].Pz +
				pBox1->Planes[i].d;

			if ( fRet < 0 ) return false;
		}
	}

	return true;
}

//=============================================================================
/**
	Ɣ̓蔻s܂

	@param pFrustum	[in] ]Rcԓ̔
	@param pBox		[in] ]Rcԓ̔

	@retval true	Փ˂Ă
	@retval false	Փ˂ĂȂ
*/
//=============================================================================
bool LunaCollision::FrustumCulling( BOX3D *pFrustum, BOX3D *pBox )
{
	Uint8 bOutside[8];

	//=========================================================================================================
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	// ݂̔̒Ɍ݂̓_Ă邩`FbN
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	//=========================================================================================================
	for ( Sint32 i = 0; i < 8; i++ )
	{
		bOutside[i] = 0x00;
		for ( Sint32 j = 0; j < 6; j++ )
		{
			Float fRet =
				pFrustum->Planes[j].a * pBox->Points[i].Px +
				pFrustum->Planes[j].b * pBox->Points[i].Py +
				pFrustum->Planes[j].c * pBox->Points[i].Pz +
				pFrustum->Planes[j].d;
			
			if ( fRet < 0 ) bOutside[i] |= (1 << j);
		}

		// |CgU̃r[tX^ʂׂĂ̓łꍇ
		// r[tX^ɂ܂
		if ( bOutside[i] == 0 ) return true;
	}

	// |CgׂĔCӂ̒P̃r[tX^ʂ̊Oɂꍇ
	// IuWFNg̓r[tX^Oɂ܂B
	if ( (bOutside[0] & bOutside[1] & bOutside[2] & bOutside[3] & bOutside[4] & bOutside[5] & bOutside[6] & bOutside[7]) != 0x00 )
	{
		return false;
	}

	//=========================================================================================================
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	//=========================================================================================================
	POINT3D Edge1[][2] = 
	{
		{ pBox->Points[0], pBox->Points[1] }, // front bottom
		{ pBox->Points[2], pBox->Points[3] }, // front top
		{ pBox->Points[0], pBox->Points[2] }, // front left
		{ pBox->Points[1], pBox->Points[3] }, // front right
		{ pBox->Points[4], pBox->Points[5] }, // back bottom
		{ pBox->Points[6], pBox->Points[7] }, // back top
		{ pBox->Points[4], pBox->Points[6] }, // back left
		{ pBox->Points[5], pBox->Points[7] }, // back right
		{ pBox->Points[0], pBox->Points[4] }, // left bottom
		{ pBox->Points[2], pBox->Points[6] }, // left top
		{ pBox->Points[1], pBox->Points[5] }, // right bottom
		{ pBox->Points[3], pBox->Points[7] }, // right top
	};
	POINT3D Face1[][4] =
	{
		{ pFrustum->Points[0], pFrustum->Points[2], pFrustum->Points[3], pFrustum->Points[1] }, // front
		{ pFrustum->Points[4], pFrustum->Points[5], pFrustum->Points[7], pFrustum->Points[6] }, // back
		{ pFrustum->Points[0], pFrustum->Points[4], pFrustum->Points[6], pFrustum->Points[2] }, // left
		{ pFrustum->Points[1], pFrustum->Points[3], pFrustum->Points[7], pFrustum->Points[5] }, // right
		{ pFrustum->Points[2], pFrustum->Points[6], pFrustum->Points[7], pFrustum->Points[3] }, // top
		{ pFrustum->Points[0], pFrustum->Points[4], pFrustum->Points[5], pFrustum->Points[1] }, // bottom
	};

	for( Sint32 i = 0; i < 12; i++ )
	{
		for( Sint32 j = 0; j < 6; j++ )
		{
			if ( EdgeIntersectsQuad( Edge1[i], Face1[j], &pFrustum->Planes[j], NULL ) )
			{
				return true;
			}
		}
	}

	//=========================================================================================================
	// r[tX^̒[̃GbŴǂꂩoEfBO{bNX̖ʂ̂̂ǂꂩɐڐG邩
	//=========================================================================================================
	POINT3D Edge2[][2] = 
	{
		{ pFrustum->Points[0], pFrustum->Points[1] }, // front bottom
		{ pFrustum->Points[2], pFrustum->Points[3] }, // front top
		{ pFrustum->Points[0], pFrustum->Points[2] }, // front left
		{ pFrustum->Points[1], pFrustum->Points[3] }, // front right
		{ pFrustum->Points[4], pFrustum->Points[5] }, // back bottom
		{ pFrustum->Points[6], pFrustum->Points[7] }, // back top
		{ pFrustum->Points[4], pFrustum->Points[6] }, // back left
		{ pFrustum->Points[5], pFrustum->Points[7] }, // back right
		{ pFrustum->Points[0], pFrustum->Points[4] }, // left bottom
		{ pFrustum->Points[2], pFrustum->Points[6] }, // left top
		{ pFrustum->Points[1], pFrustum->Points[5] }, // right bottom
		{ pFrustum->Points[3], pFrustum->Points[7] }, // right top
	};
	POINT3D Face2[][4] =
	{
		{ pBox->Points[0], pBox->Points[2], pBox->Points[3], pBox->Points[1] }, // front
		{ pBox->Points[4], pBox->Points[5], pBox->Points[7], pBox->Points[6] }, // back
		{ pBox->Points[0], pBox->Points[4], pBox->Points[6], pBox->Points[2] }, // left
		{ pBox->Points[1], pBox->Points[3], pBox->Points[7], pBox->Points[5] }, // right
		{ pBox->Points[2], pBox->Points[6], pBox->Points[7], pBox->Points[3] }, // top
		{ pBox->Points[0], pBox->Points[4], pBox->Points[5], pBox->Points[1] }, // bottom
	};

	for( Sint32 i = 0; i < 12; i++ )
	{
		for( Sint32 j = 0; j < 6; j++ )
		{
			if ( EdgeIntersectsQuad( Edge2[i], Face2[j], &pBox->Planes[j], NULL ) )
			{
				return true;
			}
		}
	}

	//=========================================================================================================
	// r[tX^oEfBO{bNXɊ܂܂邩ۂ
	// r[tX^̑S_oEfBO{bNX̊Oɂꍇ
	// r[tX^̓oEfBO{bNXɂ͊܂܂܂
	//=========================================================================================================
	for ( Sint32 i = 0; i < 6; i++ )
	{
		Float fRet =
			pFrustum->Planes[i].a * pBox->Points[0].Px +
			pFrustum->Planes[i].b * pBox->Points[0].Py +
			pFrustum->Planes[i].c * pBox->Points[0].Pz +
			pFrustum->Planes[i].d;

		if ( fRet < 0 ) return false;
	}

	return true;
}

//=============================================================================
/**
	LunaCollisionNXŎgp锠f[^̍쐬s܂

	@param pOut		[out] f[^i[
	@param pMinPt	[in] ̃oEfBO{bNXŏ_
	@param pMaxPt	[in] ̃oEfBO{bNXő_
	@param pMatrix	[in] ʏ̔Cӂ̓_
*/
//=============================================================================
void LunaCollision::CreateBox( BOX3D *pOut, POINT3D *pMinPt, POINT3D *pMaxPt, LMATRIX *pMatrix )
{
	LVECTOR3 LocalBB[] = {
		LVECTOR3( pMinPt->Px, pMinPt->Py, pMinPt->Pz ), // xyz
		LVECTOR3( pMaxPt->Px, pMinPt->Py, pMinPt->Pz ), // Xyz
		LVECTOR3( pMinPt->Px, pMaxPt->Py, pMinPt->Pz ), // xYz
		LVECTOR3( pMaxPt->Px, pMaxPt->Py, pMinPt->Pz ), // XYz
		LVECTOR3( pMinPt->Px, pMinPt->Py, pMaxPt->Pz ), // xyZ
		LVECTOR3( pMaxPt->Px, pMinPt->Py, pMaxPt->Pz ), // XyZ
		LVECTOR3( pMinPt->Px, pMaxPt->Py, pMaxPt->Pz ), // xYZ
		LVECTOR3( pMaxPt->Px, pMaxPt->Py, pMaxPt->Pz ), // XYZ
	};

	// [JWn烏[hWnɃoEfBO{bNXgXtH[܂
	((LVECTOR3*)(&(pOut->Points[0])))->TransformCoord( &LocalBB[0], pMatrix );
	((LVECTOR3*)(&(pOut->Points[1])))->TransformCoord( &LocalBB[1], pMatrix );
	((LVECTOR3*)(&(pOut->Points[2])))->TransformCoord( &LocalBB[2], pMatrix );
	((LVECTOR3*)(&(pOut->Points[3])))->TransformCoord( &LocalBB[3], pMatrix );
	((LVECTOR3*)(&(pOut->Points[4])))->TransformCoord( &LocalBB[4], pMatrix );
	((LVECTOR3*)(&(pOut->Points[5])))->TransformCoord( &LocalBB[5], pMatrix );
	((LVECTOR3*)(&(pOut->Points[6])))->TransformCoord( &LocalBB[6], pMatrix );
	((LVECTOR3*)(&(pOut->Points[7])))->TransformCoord( &LocalBB[7], pMatrix );

    // oEfBO{bNX̖ʂ𐶐܂
	CreatePlaneFromPoint( &pOut->Planes[0], &pOut->Points[0], &pOut->Points[1], &pOut->Points[2] ); // Near
    CreatePlaneFromPoint( &pOut->Planes[1], &pOut->Points[6], &pOut->Points[7], &pOut->Points[5] ); // Far
    CreatePlaneFromPoint( &pOut->Planes[2], &pOut->Points[2], &pOut->Points[6], &pOut->Points[4] ); // Left
    CreatePlaneFromPoint( &pOut->Planes[3], &pOut->Points[7], &pOut->Points[3], &pOut->Points[5] ); // Right
    CreatePlaneFromPoint( &pOut->Planes[4], &pOut->Points[2], &pOut->Points[3], &pOut->Points[6] ); // Top
    CreatePlaneFromPoint( &pOut->Planes[5], &pOut->Points[1], &pOut->Points[0], &pOut->Points[4] ); // Bottom
}

//=============================================================================
/**
	LunaCollisionNXŎgp锠f[^̍쐬s܂<BR>
	vWFNVlăf[^쐬ŝŁA<BR>
	r[tX^Ȃǂ̐ɓKĂ܂B

	@param pOut		[out] f[^i[
	@param pMinPt	[in] ̃oEfBO{bNXŏ_
	@param pMaxPt	[in] ̃oEfBO{bNXő_
	@param pMatrix	[in] ʏ̔Cӂ̓_
*/
//=============================================================================
void LunaCollision::CreateBoxProjection( BOX3D *pOut, POINT3D *pMinPt, POINT3D *pMaxPt, LMATRIX *pMatrix )
{
	LVECTOR3 LocalBB[] = {
		LVECTOR3( pMinPt->Px, pMinPt->Py, pMinPt->Pz ), // xyz
		LVECTOR3( pMaxPt->Px, pMinPt->Py, pMinPt->Pz ), // Xyz
		LVECTOR3( pMinPt->Px, pMaxPt->Py, pMinPt->Pz ), // xYz
		LVECTOR3( pMaxPt->Px, pMaxPt->Py, pMinPt->Pz ), // XYz
		LVECTOR3( pMinPt->Px, pMinPt->Py, pMaxPt->Pz ), // xyZ
		LVECTOR3( pMaxPt->Px, pMinPt->Py, pMaxPt->Pz ), // XyZ
		LVECTOR3( pMinPt->Px, pMaxPt->Py, pMaxPt->Pz ), // xYZ
		LVECTOR3( pMaxPt->Px, pMaxPt->Py, pMaxPt->Pz ), // XYZ
	};

	// [JWn烏[hWnɃoEfBO{bNXgXtH[܂
	((LVECTOR3*)(&(pOut->Points[0])))->TransformProjection( &LocalBB[0], pMatrix );
	((LVECTOR3*)(&(pOut->Points[1])))->TransformProjection( &LocalBB[1], pMatrix );
	((LVECTOR3*)(&(pOut->Points[2])))->TransformProjection( &LocalBB[2], pMatrix );
	((LVECTOR3*)(&(pOut->Points[3])))->TransformProjection( &LocalBB[3], pMatrix );
	((LVECTOR3*)(&(pOut->Points[4])))->TransformProjection( &LocalBB[4], pMatrix );
	((LVECTOR3*)(&(pOut->Points[5])))->TransformProjection( &LocalBB[5], pMatrix );
	((LVECTOR3*)(&(pOut->Points[6])))->TransformProjection( &LocalBB[6], pMatrix );
	((LVECTOR3*)(&(pOut->Points[7])))->TransformProjection( &LocalBB[7], pMatrix );

    // oEfBO{bNX̖ʂ𐶐܂
	CreatePlaneFromPoint( &pOut->Planes[0], &pOut->Points[0], &pOut->Points[1], &pOut->Points[2] ); // Near
    CreatePlaneFromPoint( &pOut->Planes[1], &pOut->Points[6], &pOut->Points[7], &pOut->Points[5] ); // Far
    CreatePlaneFromPoint( &pOut->Planes[2], &pOut->Points[2], &pOut->Points[6], &pOut->Points[4] ); // Left
    CreatePlaneFromPoint( &pOut->Planes[3], &pOut->Points[7], &pOut->Points[3], &pOut->Points[5] ); // Right
    CreatePlaneFromPoint( &pOut->Planes[4], &pOut->Points[2], &pOut->Points[3], &pOut->Points[6] ); // Top
    CreatePlaneFromPoint( &pOut->Planes[5], &pOut->Points[1], &pOut->Points[0], &pOut->Points[4] ); // Bottom
}

//=============================================================================
/**
	LunaCollisionNXŎgp镽ʃf[^̍쐬s܂

	@param pOut		[out] ʃf[^i[
	@param pPt0		[in] ʏ̔Cӂ̓_
	@param pPt1		[in] ʏ̔Cӂ̓_
	@param pPt2		[in] ʏ̔Cӂ̓_
*/
//=============================================================================
void LunaCollision::CreatePlaneFromPoint( PLANE3D *pOut, POINT3D *pPt0, POINT3D *pPt1, POINT3D *pPt2 )
{
    LVECTOR3 Edge1 = LVECTOR3(pPt0->Px - pPt1->Px, pPt0->Py - pPt1->Py, pPt0->Pz - pPt1->Pz);
    LVECTOR3 Edge2 = LVECTOR3(pPt0->Px - pPt2->Px, pPt0->Py - pPt2->Py, pPt0->Pz - pPt2->Pz);

	LVECTOR3 Normal;
	Normal.CrossProduct( &Edge1, &Edge2 );
	Normal.Normalize();

	pOut->a = Normal.x;
	pOut->b = Normal.y;
	pOut->c = Normal.z;
	pOut->d = -((pPt0->Px * Normal.x) + (pPt0->Py * Normal.y) + (pPt0->Pz * Normal.z));
}

//=============================================================================
/**
	LunaCollisionNXŎgp镽ʃf[^̍쐬s܂

	@param pOut		[out] ʃf[^i[
	@param pPt		[in] ʏ̔Cӂ̓_
	@param pNormal	[in] ʂ̖@
*/
//=============================================================================
void LunaCollision::CreatePlane( PLANE3D *pOut, POINT3D *pPt, POINT3D *pNormal )
{
	pOut->a = pNormal->Px;
	pOut->b = pNormal->Py;
	pOut->c = pNormal->Pz;
	pOut->d = -((pPt->Px * pNormal->Px) + (pPt->Py * pNormal->Py) + (pPt->Pz * pNormal->Pz));
}
}//eon
