
/** @file
    @brief		ZpNX
    @author		t`
    @since		2003.07.01
*/


//=============================================================================
// INCLUDE
//=============================================================================
#include "LunaBase.h"

//#include "LunaMain.h"
#include "LunaMath.h"

namespace lunadxmath{
//=============================================================================
// VARIABLE
//=============================================================================
Float *LunaMath::m_pfSinTable;
Uint32 *LunaMath::m_pfArcTanTable;


//=============================================================================
/**
	NX̏܂

	@retval true		
	@retval false		s
*/
//=============================================================================
bool LunaMath::Initialize( void )
{
	//Luna::LogoutMainCategory( "LunaMathNX̏" );
	//Luna::LogoutNextLine();

	//----------------------------------------------------------
	// TCe[u
	//----------------------------------------------------------
	m_pfSinTable = (Float*)MemoryAlloc( sizeof(Float[SIN_TABLE_SIZE]) );
	if ( m_pfSinTable == NULL ) return false;

	for ( Sint32 i = 0; i < SIN_TABLE_SIZE; i++ )
	{
		m_pfSinTable[i] = (Float)( sin( F(i) * PI2 / F(SIN_TABLE_SIZE) ) );
	}

	//----------------------------------------------------------
	// A[N^WFge[u
	//----------------------------------------------------------
	m_pfArcTanTable = (Uint32*)MemoryAlloc( sizeof(Uint32[ARCTAN_TABLE_SIZE+1]) );
	if ( m_pfArcTanTable == NULL ) return false;

	for ( Sint32 i = 0; i < ARCTAN_TABLE_SIZE+1; i++ )
	{
		m_pfArcTanTable[i] = (Sint32)(atan( (Float64)i * (1.0 / (Float64)ARCTAN_TABLE_SIZE) ) * (Float64)SIN_TABLE_SIZE * (1.0 / (Float64)PI2));
	}

	//----------------------------------------------------------
	// ԂŗV[h
	//----------------------------------------------------------
	RandSeed( (Uint32)time( NULL ) );

	//Luna::LogoutNextLine();

	return true;
}

//=============================================================================
/**
	ZpNX̉܂B
*/
//=============================================================================
void LunaMath::Finalize( void )
{
	//Luna::LogoutMainCategory( "LunaMathNX̉" );
	//Luna::LogoutNextLine();

	MemoryFree( m_pfSinTable );
	MemoryFree( m_pfArcTanTable );

	//Luna::LogoutNextLine();
}

//=============================================================================
/**
	TCl擾܂B

	@param Angle	[in] pxiPUTTRUj

	@return TCl
*/
//=============================================================================
Float LunaMath::Sin( Sint32 Angle )
{
	return m_pfSinTable[NORMALIZE_ANGLE(Angle)];
}

//=============================================================================
/**
	RTCl擾܂B

	@param Angle	[in] pxiPUTTRUj

	@return RTCl
*/
//=============================================================================
Float LunaMath::Cos( Sint32 Angle )
{
	Angle += SIN_TABLE_SIZE >> 2;
	return m_pfSinTable[NORMALIZE_ANGLE(Angle)];
}

//=============================================================================
/**
	TClRTCl𓯎Ɏ擾܂B

	@param Angle	[in] px
	@param pfSin	[out] TCli[|C^
	@param pfCos	[out] RTCli[|C^
*/
//=============================================================================
void LunaMath::SinCos( Sint32 Angle, Float *pfSin, Float *pfCos )
{
	*pfSin = LunaMath::Sin( Angle );
	*pfCos = LunaMath::Cos( Angle );
}

//=============================================================================
/**
	CӍWԂ̊px擾܂

	@param Dx	[in] w
	@param Dy	[in] x

	@return		px
*/
//=============================================================================
Sint32 LunaMath::Atan( Sint32 Dx, Sint32 Dy )
{
	//=================================================
	// x0̎
	//=================================================
	if ( Dx == 0 )
	{
		if ( Dy < 0 )
		{
			return ARCTAN_POS_3;
		}
		else
		{
			return ARCTAN_POS_1;
		}
	}
	//=================================================
	// y0̎
	//=================================================
	if ( Dy == 0 )
	{
		if ( Dx < 0 )
		{
			return ARCTAN_POS_2;
		}
		else
		{
			return 0;
		}
	}

	//=================================================
	// x}CiX(2/3ی)
	//=================================================
	if ( Dx < 0 )
	{
		Dx = -Dx;
		//=================================================
		// y}CiX(3ی)
		//=================================================
		if ( Dy < 0 )
		{
			Dy = -Dy;
			//=================================================
			// 180`224x
			//=================================================
			if ( Dx >= Dy )	
			{
				return ARCTAN_POS_2 + m_pfArcTanTable[(Dy << ARCTAN_POS_SHIFT) / Dx];
			}
			//=================================================
			// 225`270x
			//=================================================
			else
			{
				return ARCTAN_POS_3 - m_pfArcTanTable[(Dx << ARCTAN_POS_SHIFT) / Dy];
			}
		}
		//=================================================
		// yvX(2ی)
		//=================================================
		else
		{
			//=================================================
			// 136`180x
			//=================================================
			if ( Dx >= Dy )	
			{
				return ARCTAN_POS_2 - m_pfArcTanTable[(Dy << ARCTAN_POS_SHIFT) / Dx];
			}
			//=================================================
			// 90`135x
			//=================================================
			else						
			{
				return ARCTAN_POS_1 + m_pfArcTanTable[(Dx << ARCTAN_POS_SHIFT) / Dy];
			}
		}
	}
	//=================================================
	// xvX(1/4ی)
	//=================================================
	else
	{
		//=================================================
		// y}CiX(4ی)
		//=================================================
		if ( Dy < 0 )
		{
			Dy = -Dy;
			//=================================================
			// 315`360x
			//=================================================
			if ( Dx >= Dy )	
			{
				return ARCTAN_POS_4 - m_pfArcTanTable[(Dy << ARCTAN_POS_SHIFT) / Dx];
			}
			//=================================================
			// 180`224x
			//=================================================
			else
			{
				return ARCTAN_POS_3 + m_pfArcTanTable[(Dx << ARCTAN_POS_SHIFT) / Dy];
			}
		}
		//=================================================
		// yvX(1ی)
		//=================================================
		else
		{
			//=================================================
			// 0`45x
			//=================================================
			if ( Dx >= Dy )
			{
				return m_pfArcTanTable[(Dy << ARCTAN_POS_SHIFT) / Dx];
			}
			//=================================================
			// 45`90x
			//=================================================
			else
			{
				return ARCTAN_POS_1 - m_pfArcTanTable[(Dx << ARCTAN_POS_SHIFT) / Dy];
			}
		}
	}
}

//=============================================================================
/**
	CӍWԂ̊px擾܂

	@param Dx	[in] w
	@param Dy	[in] x

	@return		px
*/
//=============================================================================
Sint32 LunaMath::Atan( Float Dx, Float Dy )
{
	//=================================================
	// x0̎
	//=================================================
	if ( Dx == 0 )
	{
		if ( Dy < 0 )
		{
			return ARCTAN_POS_3;
		}
		else
		{
			return ARCTAN_POS_1;
		}
	}
	//=================================================
	// y0̎
	//=================================================
	if ( Dy == 0 )
	{
		if ( Dx < 0 )
		{
			return ARCTAN_POS_2;
		}
		else
		{
			return 0;
		}
	}

	//=================================================
	// x}CiX(2/3ی)
	//=================================================
	if ( Dx < 0 )
	{
		Dx = -Dx;
		//=================================================
		// y}CiX(3ی)
		//=================================================
		if ( Dy < 0 )
		{
			Dy = -Dy;
			//=================================================
			// 180`224x
			//=================================================
			if ( Dx >= Dy )	
			{
				return ARCTAN_POS_2 + m_pfArcTanTable[L(ARCTAN_SCALE(Dy) / Dx)];
			}
			//=================================================
			// 225`270x
			//=================================================
			else
			{
				return ARCTAN_POS_3 - m_pfArcTanTable[L(ARCTAN_SCALE(Dx) / Dy)];
			}
		}
		//=================================================
		// yvX(2ی)
		//=================================================
		else
		{
			//=================================================
			// 136`180x
			//=================================================
			if ( Dx >= Dy )	
			{
				return ARCTAN_POS_2 - m_pfArcTanTable[L(ARCTAN_SCALE(Dy) / Dx)];
			}
			//=================================================
			// 90`135x
			//=================================================
			else						
			{
				return ARCTAN_POS_1 + m_pfArcTanTable[L(ARCTAN_SCALE(Dx) / Dy)];
			}
		}
	}
	//=================================================
	// xvX(1/4ی)
	//=================================================
	else
	{
		//=================================================
		// y}CiX(4ی)
		//=================================================
		if ( Dy < 0 )
		{
			Dy = -Dy;
			//=================================================
			// 315`360x
			//=================================================
			if ( Dx >= Dy )	
			{
				return ARCTAN_POS_4 - m_pfArcTanTable[L(ARCTAN_SCALE(Dy) / Dx)];
			}
			//=================================================
			// 180`224x
			//=================================================
			else
			{
				return ARCTAN_POS_3 + m_pfArcTanTable[L(ARCTAN_SCALE(Dx) / Dy)];
			}
		}
		//=================================================
		// yvX(1ی)
		//=================================================
		else
		{
			//=================================================
			// 0`45x
			//=================================================
			if ( Dx >= Dy )
			{
				return m_pfArcTanTable[L(ARCTAN_SCALE(Dy) / Dx)];
			}
			//=================================================
			// 45`90x
			//=================================================
			else
			{
				return ARCTAN_POS_1 - m_pfArcTanTable[L(ARCTAN_SCALE(Dx) / Dy)];
			}
		}
	}
}

//=============================================================================
/**
	CӍW̃|Cg_Sɉ]܂

	@param pPoint	[in/out] ]|Cg
	@param Angle	[in] ]px
*/
//=============================================================================
void LunaMath::RotationPoint( LUNAPOINT *pPoint, Sint32 Angle )
{
	Float fSin = LunaMath::Sin( Angle );
	Float fCos = LunaMath::Cos( Angle );

	Float fPx = (fCos * pPoint->Px) - (fSin * pPoint->Py);
	Float fPy = (fSin * pPoint->Px) + (fCos * pPoint->Py);

	pPoint->Px = fPx;
	pPoint->Py = fPy;
}

//=============================================================================
/**
	Cӂ̒lŃV[h܂B

	@param Seed		[in] ̎
*/
//=============================================================================
void LunaMath::RandSeed( Uint32 Seed )
{
	srand( Seed );
}

//=============================================================================
/**
	CӍW̃|Cg_Sɉ]܂

	@return 
*/
//=============================================================================
Sint32 LunaMath::Rand( void )
{
	return (rand() * rand());
}

//=============================================================================
/**
	CӍW̃|Cg_Sɉ]܂

	@param Start	[in] n_
	@param End		[in] I_

	@return		
*/
//=============================================================================
Sint32 LunaMath::Rand( Sint32 Start, Sint32 End )
{
	return (Rand() % (End - Start + 1)) + Start;
}

//=============================================================================
/**
	0.0`1.0܂ł̗擾܂B

	@return		
*/
//=============================================================================
Float LunaMath::RandF( void )
{
	return (F(rand()) / F(RAND_MAX));
}

//=============================================================================
/**
	w͈̗͂擾܂B

	@param Start	[in] n_
	@param End		[in] I_

	@return		Zoꂽ
*/
//=============================================================================
Float LunaMath::RandF( Float Start, Float End )
{
	return (RandF() * (End - Start)) + Start;
}

}//eon