


//==========================================================================
// INCLUDE
//==========================================================================
#include "LunaBase.h"

#include "LunaQuaternion.h"

namespace lunadxmath{
/**
*/
void CQuaternion::Identity( void )
{
	x = 0;
	y = 0;
	z = 0;
	w = 1;
}

/**
*/
Float CQuaternion::LengthSq( void )
{
    return (x*x + y*y + z*z + w*w);
}

/**
*/
Float CQuaternion::Length( void )
{
    return (Float)sqrt( LengthSq() );
}

/**
*/
Float CQuaternion::DotProduct( CQuaternion *pIn )
{
    return ((x * pIn->x) + (y * pIn->y) + (z * pIn->z) + (w * pIn->w));
}

/**
*/
void CQuaternion::Conjugate( void )
{
    x = -x;
    y = -y;
    z = -z;
}

/**
*/
void CQuaternion::Add( CQuaternion *pIn1, CQuaternion *pIn2 )
{
	x = pIn1->x + pIn2->x;
	y = pIn1->y + pIn2->y;
	z = pIn1->z + pIn2->z;
	w = pIn1->w + pIn2->w;
}

/**
*/
void CQuaternion::Sub( CQuaternion *pIn1, CQuaternion *pIn2 )
{
	x = pIn1->x - pIn2->x;
	y = pIn1->y - pIn2->y;
	z = pIn1->z - pIn2->z;
	w = pIn1->w - pIn2->w;
}

/**
*/
void CQuaternion::Mul( CQuaternion *pIn1, CQuaternion *pIn2 )
{
	x = (pIn1->w*pIn2->x) + (pIn2->w*pIn1->x) + (pIn1->y*pIn2->z) - (pIn1->z*pIn2->y);
	y = (pIn1->w*pIn2->y) + (pIn2->w*pIn1->y) + (pIn1->z*pIn2->x) - (pIn1->x*pIn2->z);
	z = (pIn1->w*pIn2->z) + (pIn2->w*pIn1->z) + (pIn1->x*pIn2->y) - (pIn1->y*pIn2->x);
	w = (pIn1->w*pIn2->w) - (pIn1->x*pIn2->x) - (pIn1->y*pIn2->y) - (pIn1->z*pIn2->z);
}

/**
*/
void CQuaternion::Lerp( CQuaternion *pIn1, CQuaternion *pIn2, Float fRate )
{
    x = pIn1->x + (fRate * (pIn2->x - pIn1->x));
    y = pIn1->y + (fRate * (pIn2->y - pIn1->y));
    z = pIn1->z + (fRate * (pIn2->z - pIn1->z));
    w = pIn1->w + (fRate * (pIn2->w - pIn1->w));
}

/**
*/
void CQuaternion::Slerp( CQuaternion *pIn1, CQuaternion *pIn2, Float fRate )
{
    CQuaternion qTemp;
    Float fDot;

	// 2̃NH[^jI̓
    fDot = pIn1->DotProduct( pIn2 );

    // fDot < 0.0f ̎]
    if ( fDot < 0.0f )
	{
        fDot = -fDot;
        qTemp.x = -pIn2->x;
        qTemp.y = -pIn2->y;
        qTemp.z = -pIn2->z;
        qTemp.w = -pIn2->w;
    }
    else
	{
		qTemp.x = pIn2->x;
		qTemp.y = pIn2->y;
		qTemp.z = pIn2->z;
		qTemp.w = pIn2->w;
	}

	// ʐ`
    if ( fDot < 0.95f )
	{
        if ( fDot >  1.0f ) fDot =  1.0f;
        if ( fDot < -1.0f ) fDot = -1.0f;

        Float angle = F(acos( fDot ));

        Float sin1 = F(sin(angle              ));
        Float sin2 = F(sin(angle *      fRate ));
        Float sin3 = F(sin(angle * (1 - fRate)));

        x = ((pIn1->x * sin3) + (qTemp.x * sin2)) / sin1;
        y = ((pIn1->y * sin3) + (qTemp.y * sin2)) / sin1;
        z = ((pIn1->z * sin3) + (qTemp.z * sin2)) / sin1;
        w = ((pIn1->w * sin3) + (qTemp.w * sin2)) / sin1;
    }
    // px͐`
    else
	{
		Lerp( pIn1, &qTemp, fRate );
	}
}

/**
*/
void CQuaternion::Add( CQuaternion *pIn )
{
	x += pIn->x;
	y += pIn->y;
	z += pIn->z;
	w += pIn->w;
}

/**
*/
void CQuaternion::Sub( CQuaternion *pIn )
{
	x -= pIn->x;
	y -= pIn->y;
	z -= pIn->z;
	w -= pIn->w;
}

/**
*/
void CQuaternion::Mul( CQuaternion *pIn )
{
	CQuaternion qTemp = *this;
	x = (qTemp.w*pIn->x) + (pIn->w*qTemp.x) + (qTemp.y*pIn->z) - (qTemp.z*pIn->y);
	y = (qTemp.w*pIn->y) + (pIn->w*qTemp.y) + (qTemp.z*pIn->x) - (qTemp.x*pIn->z);
	z = (qTemp.w*pIn->z) + (pIn->w*qTemp.z) + (qTemp.x*pIn->y) - (qTemp.y*pIn->x);
	w = (qTemp.w*pIn->w) - (qTemp.x*pIn->x) - (qTemp.y*pIn->y) - (qTemp.z*pIn->z);
}

/**
*/
void CQuaternion::Lerp( CQuaternion *pIn, Float fRate )
{
    x += fRate * (pIn->x - x);
    y += fRate * (pIn->y - y);
    z += fRate * (pIn->z - z);
    w += fRate * (pIn->w - w);
}

/**
*/
void CQuaternion::Slerp( CQuaternion *pIn, Float fRate )
{
    CQuaternion In1 = *this;
    CQuaternion qTemp;
    Float fDot;

	// 2̃NH[^jI̓
    fDot = In1.DotProduct( pIn );

    // fDot < 0.0f ̎]
    if ( fDot < 0.0f )
	{
        fDot = -fDot;
        qTemp.x = -pIn->x;
        qTemp.y = -pIn->y;
        qTemp.z = -pIn->z;
        qTemp.w = -pIn->w;
    }
    else
	{
		qTemp.x = pIn->x;
		qTemp.y = pIn->y;
		qTemp.z = pIn->z;
        qTemp.w = pIn->w;
	}

	// ʐ`
    if ( fDot < 0.95f )
	{
        if ( fDot >  1.0f ) fDot =  1.0f;
        if ( fDot < -1.0f ) fDot = -1.0f;

        Float angle = F(acos( fDot ));

        Float sin1 = F(sin(angle              ));
        Float sin2 = F(sin(angle *      fRate ));
        Float sin3 = F(sin(angle * (1 - fRate)));

        x = ((In1.x * sin3) + (qTemp.x * sin2)) / sin1;
        y = ((In1.y * sin3) + (qTemp.y * sin2)) / sin1;
        z = ((In1.z * sin3) + (qTemp.z * sin2)) / sin1;
        w = ((In1.w * sin3) + (qTemp.w * sin2)) / sin1;
    }
    // px͐`
    else
	{
		Lerp( pIn, &qTemp, fRate );
	}
}

}