#ifndef LUNA_DX_MATH_VECTOR_HPP
#define LUNA_DX_MATH_VECTOR_HPP




namespace lunadxmath{

template<class T>
struct outer_product{
	void operator()(T *dest,T &v0,T &v2)
	{
		T t;
		t.x = v0.y*v1.z - v0.z*v1.y;
		t.y = v0.z*v1.x - v0.x*v1.z;
		t.z = v0.x*v1.y - v0.y*v1.x;
		*dest = t;
	}
};


template<class T>
class calc_vertical_vector{
	outer_product<T> func;
	void operator()(T *dest,T &v0,T &v1,T &v2){
		T t1,t2;
		
		t1 = v1 - v0;
		t2 = v2 - v0;
		func(&t2,t1,t2);
		func(dest,t1,t2);
	}
};
template<class T>
struct inner_product{
	double operator()(T &v0,T &v1)
	{
		return v0.x*v1.x + v0.y*v1.y + v0.z*v1.z;
	}
};
template<class T>
struct calc_length{
	calc_vertical_vector<T> cvv;
	double operator()(T *a,T *b,T *c)
	{
		T t1,t2;
		
		cvv(a,b,c,&t1);
		t1.normalize();
		t2 = a - c;
		return inner_product(&t1,&t2);
	}
};


}//eon

#include "vector3.hpp"
#include "vector2.hpp"
#include "vector4.hpp"

#endif